/* SuSE secure compartment v0.5 (c) 2000 by Marc Heuse <marc@suse.de>
 *
 * Description:
 * Compartment provides all possibilities to securely run insecure and/or
 * untrusted programs. It provides you with all necessary options to fine
 * grain the tightening process as you need it.
 *
 * Please note that you need Linux kernel 2.2.12 or later!
 *
 * Official releases and updates can be found on the SuSE Linux distribution
 * from version 7.0 on and ftp.suse.com plus mirror sites.
 *
 * Unofficial updates and betas can be found at http://www.suse.de/~marc
 *
 * Please contact me directly at marc@suse.de for bugs, comments or ideas.
 *
 */


#include <stdio.h>
#include <unistd.h>
#include <grp.h>
#include <pwd.h>
#include <sys/types.h>
#include <fcntl.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <linux/capability.h>
#include <stdarg.h>
#include <syslog.h>

#define PROGRAM_NAME	"SuSE secure compartment"
#define VERSION		"v0.5"
#define AUTHOR		"Marc Heuse <marc@suse.de>"
#define POINTER		"http://www.suse.de/~marc"

char *_env[] = { "HOME=/", "COMPARTMENT=YES", "PATH=/bin:/usr/bin:/", "" };

int cap_set_no[29] = {
  CAP_CHOWN,CAP_DAC_OVERRIDE,CAP_DAC_READ_SEARCH,CAP_FOWNER,CAP_FSETID,
  CAP_FS_MASK,CAP_KILL,CAP_SETGID,CAP_SETUID,CAP_SETPCAP,CAP_LINUX_IMMUTABLE,
  CAP_NET_BIND_SERVICE,CAP_NET_BROADCAST,CAP_NET_ADMIN,CAP_NET_RAW,CAP_IPC_LOCK,
  CAP_IPC_OWNER,CAP_SYS_MODULE,CAP_SYS_RAWIO,CAP_SYS_CHROOT,CAP_SYS_PTRACE,
  CAP_SYS_PACCT,CAP_SYS_ADMIN,CAP_SYS_BOOT,CAP_SYS_NICE,CAP_SYS_RESOURCE,
  CAP_SYS_TIME,CAP_SYS_TTY_CONFIG, 0 };
char cap_set_names[29][29] = {
  "CAP_CHOWN","CAP_DAC_OVERRIDE","CAP_DAC_READ_SEARCH","CAP_FOWNER","CAP_FSETID",
  "CAP_FS_MASK","CAP_KILL","CAP_SETGID","CAP_SETUID","CAP_SETPCAP",
  "CAP_LINUX_IMMUTABLE","CAP_NET_BIND_SERVICE","CAP_NET_BROADCAST",
  "CAP_NET_ADMIN","CAP_NET_RAW","CAP_IPC_LOCK","CAP_IPC_OWNER","CAP_SYS_MODULE",
  "CAP_SYS_RAWIO","CAP_SYS_CHROOT","CAP_SYS_PTRACE","CAP_SYS_PACCT",
  "CAP_SYS_ADMIN","CAP_SYS_BOOT","CAP_SYS_NICE","CAP_SYS_RESOURCE","CAP_SYS_TIME",
  "CAP_SYS_TTY_CONFIG", "" };

extern int capset(cap_user_header_t header, cap_user_data_t data);
extern char **environ;
int do_chroot, do_group, do_user, do_caps, do_init, program_params, verbose, quiet;
char *chroot_path, *init_program, *prg, *user, **_argv;
long tmp;
__u32 temp, caps = 0;
uid_t set_user;
gid_t set_group;
struct passwd *pw;

void help() {
    fprintf(stderr, "%s %s %s %s\n\n", PROGRAM_NAME, VERSION, AUTHOR, POINTER);
    fprintf(stderr, "Syntax: %s [options] /full/path/to/program\n\n", prg);
    fprintf(stderr, "Options:
\t --chroot path\t chroot to path
\t --user user\t change uid to this user
\t --group group\t change gid to this group
\t --init program\t execute this program/script before doing anything
\t --cap capset\t set capset name. You can specify several capsets.
\t --verbose\t be verbose
\t --quiet\t do no logging (to syslog)
\nI know the following capset names:");
    tmp = 0;
    while(strlen(cap_set_names[tmp])>0)
        fprintf(stderr," %s",cap_set_names[tmp++]);
    fprintf(stderr, " (see linux/capability.h for more information)\n\n");
    exit(-1);
}

void print_msg(const char *format, ...) {
    va_list arg;
    char log[4096];
    if (quiet == 0) {
        va_start(arg, format);
        vfprintf(stderr, format, arg);
        vsnprintf(log, sizeof(log), format, arg);
        syslog(LOG_NOTICE, log);
        va_end(arg);
    }
}

void my_secure() {
    char file[10] = "/dev/null";
    int fd;

    environ = _env;             // set a safe and clean environement
    if (verbose)
        print_msg("Safe environment set\n");
    if ((fd = open(file, O_RDWR)) < 0) {
	file[1] = '\0';
	if ((fd = open(file, O_RDWR)) < 0) {
	    print_msg("Can not open /dev/null nor /, no more fd's?\n");
	    exit(-1);
	}
    }
    while (fd < 2)
        fd = open(file, O_RDWR);
    close(fd);
    if (verbose)
        print_msg("Ensured that fd 0-2 are open\n");

    /* FIXME: should limits be reset or signals be blocked or ignored? */

    return;
}

void my_caps() {
    cap_user_header_t head = (cap_user_header_t) calloc(1, sizeof(cap_user_header_t));
    cap_user_data_t cap = (cap_user_data_t)   calloc(1, sizeof(cap_user_data_t));
    head->version = _LINUX_CAPABILITY_VERSION;
    head->pid = 0;
    cap->inheritable = cap->permitted = cap->effective = caps;
    if (capset(head, cap) != 0) {
        print_msg("Could not set capabilities.\n");
        exit(-1);
    } else
        if (verbose)
            print_msg("Capabilities successfully set to 0x%0x\n",cap->effective);
    free(head);
    free(cap);
    return;
}
 
int main (int argc, char *argv[]) {
    struct stat st;
/* process the parameters */
    prg = argv[0];
    if (argc < 2)
        help();
    if (strcmp(argv[1],"-h") == 0)
        help();
    program_params = 1;
    openlog("SuSEcompartment", LOG_PID, LOG_DAEMON);
    while((argc - 1) > program_params && strncmp(argv[program_params], "--", 2) == 0) {
        if (strcmp(argv[program_params], "--chroot") == 0) {
            chroot_path = argv[++program_params];
            do_chroot = 1;
        } else
        if (strcmp(argv[program_params], "--user") == 0) {
            if ((pw = (struct passwd *) getpwnam(argv[++program_params])) != NULL) {
                user = argv[program_params];
                set_user = pw->pw_uid;
            } else {
                tmp = strtol(argv[program_params], (char **)NULL, 10);
                if (tmp < 0 || tmp > 65535) {
                    print_msg("--user is out of bounds (username or a value between 0 and 65535): %ld\n\n", tmp);
                    exit(-1);
                }
                set_user = (uid_t) tmp;
            }
            if (verbose)
                print_msg("UID will be set to %d\n", set_user);
            do_user = 1;
        } else
        if (strcmp(argv[program_params], "--group") == 0) {
            if ((pw = (struct passwd *) getpwnam(argv[++program_params])) != NULL) {
                set_user = pw->pw_uid;
            } else {
                tmp = strtol(argv[program_params], (char **)NULL, 10);
                if (tmp < 0 || tmp > 65535) {
                    print_msg("--group is out of bounds (groupname or a value between 0 and 65535): %ld\n\n", tmp);
                    exit(-1);
                }
                set_group = (gid_t) tmp;
            }
            if (verbose)
                print_msg("GID will be set to %d\n", set_group);
            do_group = 1;
        } else
        if (strcmp(argv[program_params], "--init") == 0) {
            init_program = argv[++program_params];
            if (verbose)
                print_msg("Initial program will be %s\n", init_program);
            do_init = 1;
        } else
        if (strcmp(argv[program_params], "--cap") == 0) {
	    program_params++;
	    tmp = 0;
	    temp = caps;
	    while((temp == caps) && (strlen(cap_set_names[tmp]) > 0)) {
		if (strcmp(argv[program_params], cap_set_names[tmp]) == 0) {
		    temp = cap_set_no[tmp];
		    caps |= (1<<((temp)&31));
		    if (verbose)
		        print_msg("Capabilities will be set to 0x%0x\n",caps);
		}
		tmp++;
	    }
	    if (caps == temp) {
		print_msg("Unknown capset name: %s\n", argv[program_params]);
		exit(-1);
	    }
            do_caps = 1;
        } else
        if (strcmp(argv[program_params], "--verbose") == 0) {
            print_msg("I am in verbose mode now\n");
            verbose = 1;
        } else
        if (strcmp(argv[program_params], "--quiet") == 0) {
            quiet = 1;
        } else
        if (strcmp(argv[program_params], "--help") == 0) {
            help();
        } else {
            print_msg("Unknown parameter: %s\n\n",argv[program_params]);
            help();
        }
        program_params++;
    }
    
/* the first parameter without an -- has to be the program plus options to execute */
    if (program_params == argc) {
        print_msg("No program given to execute.\n\n");
        help();    
    }
    _argv = argv;
    _argv += program_params;

/* some further insightful deductions */
    if (do_user == 0)
        set_user = geteuid();
    if (do_group == 0)
        set_group = getegid();
    if (user == NULL) {
        if ((pw = (struct passwd *) getpwuid(set_user)) == NULL) {
            print_msg("Warning: username for uid %u is not found in /etc/passwd.\n", set_user);
        } else
            user = (char *) pw->pw_name;
    }

    if (do_caps && do_user) {
	print_msg("Warning: --user and --caps will not work together!\n");
    }

    my_secure();

    if (do_init)
        if (system(init_program)) {
            print_msg("Error executing init program %s\n", init_program);
            exit(1);
        } else
            if (verbose)
                print_msg("Init program %s run successfully\n", init_program);

    if (do_chroot) {
        if (chroot(chroot_path)) {
            print_msg("Error chrooting to %s\n", chroot_path); 
            exit(1);
        }
        chdir("/");
        if (verbose)
            print_msg("Chrooted sucessfully to %s\n", chroot_path);
    }

    if (do_group)
        if (setgid(set_group)) {
            print_msg("Error setting gid to %u\n", set_group);
            exit(1);
        } else
            if (verbose)
                print_msg("GID successfully set to %d\n",set_group);

    if (do_user || do_group) {
        if (user != NULL) {
            if (initgroups(user, set_group)) {
                print_msg("Error setting initgroups.\n");
                exit(1);
            } else
                if (verbose)
                    print_msg("Initgroups successfully set\n");
        } else
            print_msg("Can not do initgroups() because username for uid %u is not found in /etc/passwd.\n", set_user);
    }

    if (do_caps)
        my_caps();

    if (do_user)
        if (setuid(set_user)) {
            print_msg("Error setting uid to %u\n", set_user);
            exit(1);
        } else
            if (verbose)
                print_msg("UID successfully set to %d\n", set_user);

    if (quiet == 0) {
        char buf[5000] = "Running as: ", buff[5000];
        int i = 1;
        strcat(buf, argv[0]); // argv[0] can not be bigger than 4096
        while (i < argc) {
	    snprintf(buff, sizeof(buf), "%s %s", buf, argv[i++]);
	    strcpy(buf, buff);
        }
        syslog(LOG_NOTICE, "%s\n", buf);
    }

    execvp(_argv[0], _argv);
    if (do_chroot == 0)
	print_msg("Could not execute %s\n", _argv[0]);
    else {
        if (stat(_argv[0], &st) < 0)
	    print_msg("Could not find file: %s\n", _argv[0]);
        else
	    if (access(_argv[0], X_OK) < 0)
		print_msg("Execute bit missing, or no permissions to execute %s\n", _argv[0]);
	    else
	        print_msg("Could not properly execute %s - the chroot environment might not be
set up correctly:
Create the directories etc and lib in (chroot_dir) and copy all necessary
libraries to (chroot_dir)/lib. Then chdir to (chroot_dir) and execute
\"ldconfig -X -r .\"\n", _argv[0]);
    }
    return 1;
}
